/** @file
 *  Certificate verification utils
 *
 *  Description: main point
 *
 *  Copyright 2017 by ISBC Group
 *  All rights reserved.
 */

#ifndef _WINDOWS
    #include <dlfcn.h>
    #include <string.h>
	#include <termios.h>
	#include <unistd.h>
#endif

#include <cryptoki.h>

#ifdef __APPLE__
    #include <PCSC/wintypes.h>
    #include <PCSC/winscard.h>
#elif _WINDOWS
	#include "stdafx.h"
#else
    #include <wintypes.h>
    #include <winscard.h>
    #include <reader.h>
#endif

#include <string>
#include <vector>
#include <iostream>
#include <fstream>
#include <sstream>
#include <time.h>
#include <algorithm>
#include <gost.h>

using namespace std;

CK_FUNCTION_LIST_PTR p11 = NULL;
bool IsInitialized = false;
	
#ifdef _WINDOWS
    HMODULE pkcsLib;
#else
    void *pkcsLib;
#endif

/**
 * cleanup allocated resources
*/
void cleanup()
{
	// Finalase library PKCS#11
	if (IsInitialized)
	{
		IsInitialized = false;
		if(p11 != NULL)
		{
			p11->C_Finalize(NULL);
			p11 = NULL;
		}
	}

	if (pkcsLib)
	{
#ifdef _WINDOWS
		FreeLibrary(pkcsLib);
#else
	    dlclose(pkcsLib);
#endif
	    pkcsLib = NULL;
	}
}

/**
 * manupulate echo to enter password secure
*/
void SetStdinEcho(bool enable = true)
{
#ifdef _WINDOWS
    HANDLE hStdin = GetStdHandle(STD_INPUT_HANDLE); 
    DWORD mode;
    GetConsoleMode(hStdin, &mode);

    if( !enable )
        mode &= ~ENABLE_ECHO_INPUT;
    else
        mode |= ENABLE_ECHO_INPUT;

    SetConsoleMode(hStdin, mode );

#else
    struct termios tty;
    tcgetattr(STDIN_FILENO, &tty);
    if( !enable )
        tty.c_lflag &= ~ECHO;
    else
        tty.c_lflag |= ECHO;

    (void) tcsetattr(STDIN_FILENO, TCSANOW, &tty);
#endif
}

/**
 * Read file as binary to BYTE buffer
*/
bool readFile(const string& path, vector<BYTE>& buffer)
{
	ifstream file(path.c_str(), ios::in|ios::binary|ios::ate);
	if (file.is_open())
	{
		streampos size = file.tellg();
		buffer.assign(static_cast< vector<BYTE>::size_type>(size), 0);
		file.seekg (0, ios::beg);
		file.read(reinterpret_cast<char *>(&buffer[0]), size);
		file.close();
		return true;
	}
	return false;
}

/**
 * Find certificate by label
 * \param hSession - session handled for opened session
 * \param label - array with label
 * \param labelLength - length of array
 * \param phObject - pointer to founded object handler (CK_INVALID_HANDLE if not found)
 * \return - result code, if 0 - success otherwise error code returned
*/
CK_RV findCertificateHandleByLabel(CK_SESSION_HANDLE hSession, CK_BYTE_PTR label, CK_ULONG labelLength, CK_OBJECT_HANDLE_PTR phObject)
{
	// Find ca cert
	CK_OBJECT_CLASS obj_class = CKO_CERTIFICATE;
    CK_CERTIFICATE_TYPE certType = CKC_X_509;
	CK_ATTRIBUTE attrTemplate[] = {
		{CKA_CLASS, &obj_class, sizeof(obj_class)},
        {CKA_CERTIFICATE_TYPE, &certType, sizeof(certType)},
		{CKA_LABEL, label, labelLength}
	};

	CK_RV rv = p11->C_FindObjectsInit(hSession, attrTemplate, 3);
	if (rv != CKR_OK)
	{
		cout << "C_FindObjectsInit failed. Code: " << rv << endl;
		return rv;
	}

	CK_ULONG ulObjectCount;
	rv = p11->C_FindObjects(hSession, phObject, 1, &ulObjectCount);
	if (rv != CKR_OK)
	{
		cout << "C_FindObjects failed. Code: " << rv << endl;
		return rv;
	}

	rv = p11->C_FindObjectsFinal(hSession);
	if (rv != CKR_OK)
	{
		cout << "C_FindObjectsFinal failed. Code: " << rv << endl;
		return rv;
	}

	return rv;
}

/**
 * Add certificate
 * \param hSession - session handled for opened session
 * \param label - array with label
 * \param labelLength - length of array
 * \param buffer - array with certificate content (DER format)
 * \param bufferLength - buffer length
 * \param phObject - pointer to founded object handler (CK_INVALID_HANDLE if not found)
 * \param isTrusted - flag which show if new certificate should be assed in trusted certificate store
 * \return - result code, if 0 - success otherwise error code returned
*/
CK_RV addNewCertificate(CK_SESSION_HANDLE hSession, CK_BYTE_PTR label, CK_ULONG labelLength,
						CK_BYTE_PTR buffer, CK_ULONG bufferLength,
						CK_OBJECT_HANDLE_PTR phObject, CK_BBOOL isTrusted)
{
	if (isTrusted)
	{
		cout << "Please enter SO PIN (Admin PIN): ";
	}
	else
	{
		cout << "Please enter PIN (User PIN): ";
	}
	
	SetStdinEcho(false);
	string pin;
	cin >> pin;
	SetStdinEcho(true);
	cout << endl;
	
	vector<BYTE> pinBuff(pin.begin(), pin.end());
	CK_RV rv = p11->C_Login(hSession, isTrusted ? CKU_SO : CKU_USER, &pinBuff[0], static_cast<ULONG>(pinBuff.size()));
	if (rv != CKR_OK && rv != CKR_USER_ALREADY_LOGGED_IN)
	{
		cout << "Login failed. Code: " << rv << endl;
		return rv;
	}

	CK_BBOOL _true = CK_TRUE;
	CK_BBOOL _false = CK_FALSE;
	CK_OBJECT_CLASS obj_class = CKO_CERTIFICATE;
    CK_CERTIFICATE_TYPE certType = CKC_X_509;
	CK_ATTRIBUTE certificateTemplate[] = {
		{CKA_CLASS, &obj_class, sizeof(obj_class)},
		{CKA_CERTIFICATE_TYPE, &certType, sizeof(certType)},
		{CKA_TOKEN, &_true, sizeof(_true)},
		{CKA_LABEL, label, labelLength},
		{CKA_ID, label, labelLength},
		{CKA_TRUSTED, (isTrusted ? &_true : &_false), sizeof(_true)},
		{CKA_VALUE, &buffer[0], bufferLength}
	};
    
	rv = p11->C_CreateObject(hSession, certificateTemplate, 7, phObject);
	if (rv != CKR_OK)
	{
		cout << "C_CreateObject for certificate failed. Code: " << rv << endl;
		return rv;
	}

	rv = p11->C_Logout(hSession);
	if (rv != CKR_OK)
	{
		cout << "Logout failed. Code: " << rv << endl;
		return rv;
	}

	return rv;
}

/**
 * Find certificate by label and add new if not found
 * \param hSession - session handled for opened session
 * \param strLabel - string with label
 * \param phObject - pointer to founded object handler (CK_INVALID_HANDLE if not found)
 * \param isTrusted - flag which show if new certificate should be assed in trusted certificate store
 * \param filePath - file path to certificate
 * \return - result code, if 0 - success otherwise error code returned
*/
CK_RV findOrAddNewCertificate(CK_SESSION_HANDLE hSession, const string& strLabel,
						CK_OBJECT_HANDLE_PTR phObject, CK_BBOOL isTrusted, const string& filePath)
{
	*phObject = CK_INVALID_HANDLE;
	vector<CK_BYTE> label(strLabel.begin(), strLabel.end());
	CK_RV rv = findCertificateHandleByLabel(
		hSession, &label[0], static_cast<CK_ULONG>(label.size()), phObject);
	if (rv != CKR_OK)
	{
		return rv;
	}

	if (CK_INVALID_HANDLE == *phObject)
	{
		// create new object
		cout << "Certificate with label " << strLabel << " was not found. Trying to load it..." << endl;

		vector<CK_BYTE> ceftBuffer;
		if (!readFile(filePath, ceftBuffer))
		{
			cout << "Can't read file: " << filePath << endl;
			return CKR_CANCEL;
		}
		rv = addNewCertificate(hSession, &label[0],  static_cast<CK_ULONG>(label.size()),
						&ceftBuffer[0], static_cast<CK_ULONG>(ceftBuffer.size()),
						phObject, isTrusted);
		if (rv != CKR_OK)
		{
			return rv;
		}
	}
	else
	{
		cout << "Certificate with label " << strLabel << " found." << endl;
	}

	return rv;
}


/**
* Verify PKCS#7 file signature with certificate chain chek via CRL files.
* \param hSession - session handled for opened session
* \return - result code, if 0 - success otherwise error code returned
*/
CK_RV pkcs7VerifyEx(CK_SESSION_HANDLE hSession)
{
	string pkcs7Path;
	vector<CK_BYTE> pkcs7Buffer;
	
	cout << "Enter PKCS#7 file path: ";	
	cin >> pkcs7Path;	

	if (!readFile(pkcs7Path, pkcs7Buffer))
	{
		cout << "Can't read file: " << pkcs7Path << endl;
		return CKR_CANCEL;
	}

	string rootCrlPath;
	vector<CK_BYTE> rootCrlBuffer;

	cout << "Enter root CA CRL file path: ";
	cin >> rootCrlPath;
	
	if (!readFile(rootCrlPath, rootCrlBuffer))
	{
		cout << "Can't read file: " << rootCrlPath << endl;
		return CKR_CANCEL;
	}

	string intCrlPath;
	vector<CK_BYTE> intCrlBuffer;

	cout << "Enter intermediate CA CRL file path: ";
	cin >> intCrlPath;
	
	if (!readFile(intCrlPath, intCrlBuffer))
	{
		cout << "Can't read file: " << intCrlPath << endl;
		return CKR_CANCEL;
	}

	const int crlCount = 2;

	CK_BYTE_PTR* crls = new CK_BYTE_PTR[crlCount]
	{
			rootCrlBuffer.data(),
			intCrlBuffer.data()
	};

	CK_ULONG_PTR crlLengths = new CK_ULONG[crlCount]
	{
			rootCrlBuffer.size(),
			intCrlBuffer.size()
	};

	CK_RV rv = p11->C_ISBC_pkcs7VerifyEx(hSession, pkcs7Buffer.data(), pkcs7Buffer.size(), NULL, 0UL, crls, crlLengths, crlCount, FALSE);

	if (rv != CKR_OK)
	{
		cout << "C_ISBC_pkcs7VerifyEx failed. Code: " << rv << endl;		
	}
	else
	{
		cout << "OK" << endl;
	}

	return rv;
}

int main(int argc, char* argv[])
{
    // uncomment to chech gost2012
    string objPrefix = "gost2012_";
    string caPath = "certs/gost2012_ca.crt";
    string userPath = "certs/gost2012_client.crt";
    string intPath = "certs/gost2012_int_valid.crt";
    string crlPath = "certs/gost2012_crl.der";
    string userByIntPath = "certs/gost2012_client_by_int.crt";
    string userByIntRevPath = "certs/gost2012_client_by_int_revoked.crt";
    string intCrlPath = "certs/gost2012_int_crl.der";
    string intRevPath = "certs/gost2012_int_revoked.crt";
    string userByRevIntPath = "certs/gost2012_client_by_revoked_int.crt";
    
    // uncomment to chech gost
	//string objPrefix = "gost_";
	//string caPath = "certs/gost_ca.crt";
	//string userPath = "certs/gost_client.crt";
	//string intPath = "certs/gost_int_valid.crt";
	//string crlPath = "certs/gost_crl.der";
	//string userByIntPath = "certs/gost_client_by_int.crt";
	//string userByIntRevPath = "certs/gost_client_by_int_revoked.crt";
	//string intCrlPath = "certs/gost_int_crl.der";
	//string intRevPath = "certs/gost_int_revoked.crt";
	//string userByRevIntPath = "certs/gost_client_by_revoked_int.crt";

	//  uncomment to chech  rsa
	//string objPrefix = "rsa_";
	//string caPath = "certs/rsa_ca.crt";
	////string userPath = "certs/rsa_client.crt"; // uncomment this to verify revoked certificate
	//string userPath = "certs/rsa_client2.crt";
	//string intPath = "";
	//string crlPath = "certs/rsa_crl.der";
	//string userByIntPath = "";
	//string userByIntRevPath = "";
	//string intCrlPath = "";
	//string intRevPath = "";
	//string userByRevIntPath = "";

	//////////////// Init library PKCS#11
#ifdef _WINDOWS
	pkcsLib = LoadLibrary("isbc_pkcs11_main.dll");
	if (!pkcsLib)
	{
		cout << "Error while load PKCS#11: "  << GetLastError() << endl;
		cleanup();
		return 1;
	}
#elif __APPLE__
    pkcsLib = dlopen("libisbc_pkcs11_main.dylib", RTLD_NOW);
#else
	pkcsLib = dlopen("libisbc_pkcs11_main.so", RTLD_NOW);

	if (!pkcsLib)
	{
		cout << "Error while load PKCS#11: "  << dlerror() << endl;
		cleanup();
		return 1;
	}
#endif
	
	void  (*pfoo)(CK_FUNCTION_LIST_PTR*);
	CK_RV rv;
	CK_ULONG ulSlotCount;
	
#ifdef _WINDOWS
	pfoo = (void (*)(CK_FUNCTION_LIST_PTR*)) GetProcAddress(pkcsLib, "C_GetFunctionList");
#else
	pfoo = (void (*)(CK_FUNCTION_LIST_PTR*)) dlsym(pkcsLib, "C_GetFunctionList");
#endif
	
	if(pfoo == NULL)
	{
		cout << "Can't get C_GetFunctionList symbol" << endl;
		cleanup();
		return 2;
	}

	pfoo(&p11);
	if(p11 == NULL)
	{
		cout << "C_GetFunctionList error" << endl;
		cleanup();
		return 3;
	}

	rv = p11->C_Initialize(NULL);
	if(rv == CKR_CRYPTOKI_ALREADY_INITIALIZED)
	{
		IsInitialized = true;
	}

	if (!IsInitialized && rv != CKR_OK)
	{
	    cout << "C_Initialize error. Code: " << rv << endl;
		cleanup();
	    return 4;
	}

	rv = p11->C_GetSlotList(FALSE, NULL_PTR, &ulSlotCount);
	if (rv != CKR_OK)
	{
	    cout << "C_GetSlotList failed. Code: " << rv << endl;
		cleanup();
	    return 5;
	}

	IsInitialized = true;

	// open session
	CK_SESSION_HANDLE	hSession;
	CK_SLOT_ID SlotId = 0;
	rv = p11->C_OpenSession(SlotId, CKF_SERIAL_SESSION | CKF_RW_SESSION, NULL, NULL, &hSession);

	if(rv != CKR_OK)
	{
		cout << "C_OpenSession failed. Code: " << rv << endl;
		cleanup();
		return 6;
	}

	// VERIFY_EX

	if (argc > 1 && string("--sig-verify-ex") == string(argv[1]))
	{
		rv = pkcs7VerifyEx(hSession);

		if (rv != CKR_OK)
		{
			cout << "pkcs7VerifyEx failed. Code: " << rv << endl;
			cleanup();
			return 100;
		}

		p11->C_CloseSession(hSession);
		cleanup();
		return 0;
	}	

	// Find ca cert
	CK_OBJECT_HANDLE hCAObject;
	string caLabel = objPrefix + "ca-certificate";
	rv = findOrAddNewCertificate(hSession, caLabel, &hCAObject, CK_TRUE, caPath);
	if (rv != CKR_OK)
	{
		p11->C_CloseSession(hSession);
		cleanup();
		return 7;
	}

	// Find user cert
	CK_OBJECT_HANDLE hUserObject = NULL;
	string userLabel = objPrefix + "user-certificate";
	rv = findOrAddNewCertificate(hSession, userLabel, &hUserObject, CK_FALSE, userPath);
	if (rv != CKR_OK)
	{
		p11->C_CloseSession(hSession);
		cleanup();
		return 8;
	}

	cout << ">>>> Start chain verification without CRL: ";
	CK_OBJECT_HANDLE trustedCertificates[] = { hCAObject };
	rv = p11->C_ISBC_CertVerify(hSession, hUserObject, trustedCertificates, 1, NULL, 0, NULL, 0, 0);
	if (rv == CKR_OK)
	{
		cout << "SUCCESS." << endl;
	}
	else if (rv == 0x80000500)
	{
		cout << "UNSUCCESS!!!" << endl;
	}
	else
	{
		cout << "C_ISBC_CertVerify failed. Code: " << rv << endl;
		p11->C_CloseSession(hSession);
		cleanup();
		return 9;
	}

	cout << "Loading CRL" << endl;
	vector<CK_BYTE> crlBuffer;
	if (!readFile(crlPath, crlBuffer))
	{
		cout << "Can't read file: " << crlPath << endl;
		return 10;
	}
	
	cout << ">>>> Start chain verification with CRL: ";
	CK_BYTE_PTR crls[] = { &crlBuffer[0] };
	CK_ULONG crlsLengths[] = { static_cast<CK_ULONG>(crlBuffer.size()) };

	rv = p11->C_ISBC_CertVerify(hSession, hUserObject, trustedCertificates, 1, NULL, 0, crls, crlsLengths, 1);
	if (rv == CKR_OK)
	{
		cout << "SUCCESS." << endl;
	}
	else if (rv == 0x80000500)
	{
		cout << "UNSUCCESS!!!" << endl;
	}
	else
	{
		cout << "C_ISBC_CertVerify failed #2. Code: " << rv << endl;
		p11->C_CloseSession(hSession);
		cleanup();
		return 11;
	}

	if (intPath.empty())
	{
		p11->C_CloseSession(hSession);
		cleanup();
		return 0;
	}

	// Find intermediate cert
	CK_OBJECT_HANDLE hIntObject = CK_INVALID_HANDLE;
	string intLabel = objPrefix + "intermediate-certificate";
	rv = findOrAddNewCertificate(hSession, intLabel, &hIntObject, CK_FALSE, intPath);
	if (rv != CKR_OK)
	{
		p11->C_CloseSession(hSession);
		cleanup();
		return 12;
	}

	// Find user cert issued by intermediate
	CK_OBJECT_HANDLE hUserIntObject = CK_INVALID_HANDLE;
	string userIntLabel = objPrefix + "user-by-intermediate-certificate";
	rv = findOrAddNewCertificate(hSession, userIntLabel, &hUserIntObject, CK_FALSE, userByIntPath);
	if (rv != CKR_OK)
	{
		p11->C_CloseSession(hSession);
		cleanup();
		return 13;
	}

	cout << ">>>> Start 2 depth chain verification without CRL: ";
	CK_OBJECT_HANDLE intermediateCertificates[] = { hIntObject };
	rv = p11->C_ISBC_CertVerify(hSession, hUserIntObject, trustedCertificates, 1, intermediateCertificates, 1, NULL, 0, 0);
	if (rv == CKR_OK)
	{
		cout << "SUCCESS." << endl;
	}
	else if (rv == 0x80000500)
	{
		cout << "UNSUCCESS!!!" << endl;
	}
	else
	{
		cout << "C_ISBC_CertVerify failed. Code: " << rv << endl;
		p11->C_CloseSession(hSession);
		cleanup();
		return 9;
	}

	cout << "Loading CRL for intermediate" << endl;
	vector<CK_BYTE> crlIntBuffer;
	if (!readFile(intCrlPath, crlIntBuffer))
	{
		cout << "Can't read file: " << intCrlPath << endl;
		return 10;
	}
	
	cout << ">>>> Start 2 depth chain verification with CRL: ";
	CK_BYTE_PTR crls2[] = { &crlBuffer[0], &crlIntBuffer[0] };
	CK_ULONG crlsLengths2[] = { static_cast<CK_ULONG>(crlBuffer.size()), static_cast<CK_ULONG>(crlIntBuffer.size()) };

	rv = p11->C_ISBC_CertVerify(hSession, hUserObject, 
		trustedCertificates, 1, intermediateCertificates, 1, crls2, crlsLengths2, 2);
	if (rv == CKR_OK)
	{
		cout << "SUCCESS." << endl;
	}
	else if (rv == 0x80000500)
	{
		cout << "UNSUCCESS!!!" << endl;
	}
	else
	{
		cout << "C_ISBC_CertVerify failed #2. Code: " << rv << endl;
		p11->C_CloseSession(hSession);
		cleanup();
		return 11;
	}

	// Find user cert issued and revoked by intermediate
	CK_OBJECT_HANDLE hUserRevokedIntObject = CK_INVALID_HANDLE;
	string userRevIntLabel = objPrefix + "user-revoked-by-intermediate-certificate";
	rv = findOrAddNewCertificate(hSession, userRevIntLabel, &hUserRevokedIntObject, CK_FALSE, userByIntRevPath);
	if (rv != CKR_OK)
	{
		p11->C_CloseSession(hSession);
		cleanup();
		return 13;
	}

	cout << ">>>> Start 2 depth chain verification for revoked certificate without CRL: ";
	rv = p11->C_ISBC_CertVerify(hSession, hUserRevokedIntObject, trustedCertificates, 1, intermediateCertificates, 1, NULL, 0, 0);
	if (rv == CKR_OK)
	{
		cout << "SUCCESS." << endl;
	}
	else if (rv == 0x80000500)
	{
		cout << "UNSUCCESS!!!" << endl;
	}
	else
	{
		cout << "C_ISBC_CertVerify failed. Code: " << rv << endl;
		p11->C_CloseSession(hSession);
		cleanup();
		return 9;
	}
	
	cout << ">>>> Start 2 depth chain verification for revoked certificate with CRL: ";
	rv = p11->C_ISBC_CertVerify(hSession, hUserRevokedIntObject, 
		trustedCertificates, 1, intermediateCertificates, 1, crls2, crlsLengths2, 2);
	if (rv == CKR_OK)
	{
		cout << "UNSUCCESS!!!" << endl;
	}
	else if (rv == 0x80000500)
	{
		cout << "SUCCESS. Certificate revoked." << endl;
	}
	else
	{
		cout << "C_ISBC_CertVerify failed #2. Code: " << rv << endl;
		p11->C_CloseSession(hSession);
		cleanup();
		return 11;
	}

	// Find revoked intermediate cert
	CK_OBJECT_HANDLE hIntRevObject = CK_INVALID_HANDLE;
	string intRevLabel = objPrefix + "revoked-intermediate-certificate";
	rv = findOrAddNewCertificate(hSession, intRevLabel, &hIntRevObject, CK_FALSE, intRevPath);
	if (rv != CKR_OK)
	{
		p11->C_CloseSession(hSession);
		cleanup();
		return 12;
	}

	// Find user cert issued by revoked intermediate
	CK_OBJECT_HANDLE hUserIntRevObject = CK_INVALID_HANDLE;
	string userIntRevLabel = objPrefix + "user-by-revoked-intermediate-certificate";
	rv = findOrAddNewCertificate(hSession, userIntRevLabel, &hUserIntRevObject, CK_FALSE, userByRevIntPath);
	if (rv != CKR_OK)
	{
		p11->C_CloseSession(hSession);
		cleanup();
		return 13;
	}

	cout << ">>>> Start 2 depth chain verification for revoked intermediate certificate without CRL: ";
	CK_OBJECT_HANDLE intermediateCertificates2[] = { hIntRevObject };
	rv = p11->C_ISBC_CertVerify(hSession, hUserIntRevObject, trustedCertificates, 1, intermediateCertificates2, 1, NULL, 0, 0);
	if (rv == CKR_OK)
	{
		cout << "SUCCESS." << endl;
	}
	else if (rv == 0x80000500)
	{
		cout << "UNSUCCESS!!!" << endl;
	}
	else
	{
		cout << "C_ISBC_CertVerify failed. Code: " << rv << endl;
		p11->C_CloseSession(hSession);
		cleanup();
		return 9;
	}
	
	cout << ">>>> Start 2 depth chain verification for revoked certificate with CRL: ";
	rv = p11->C_ISBC_CertVerify(hSession, hUserIntRevObject, 
		trustedCertificates, 1, intermediateCertificates2, 1, crls, crlsLengths, 1);
	if (rv == CKR_OK)
	{
		cout << "UNSUCCESS!!!" << endl;
	}
	else if (rv == 0x80000500)
	{
		cout << "SUCCESS. Certificate revoked." << endl;
	}
	else
	{
		cout << "C_ISBC_CertVerify failed #2. Code: " << rv << endl;
		p11->C_CloseSession(hSession);
		cleanup();
		return 11;
	}

    p11->C_CloseSession(hSession);
	cleanup();
}
