/**
 *  ESMART Token v3.1 - PKCS#11 Sample
 *
 *  Description: CBasicOperator realization
 *
 *  $Author: Alexander Revo
 *  $Date: 14.04.2014
 *  $Id: 3513-02-011
 *
 *  Copyright 2014 by ISBC Group
 *  All rights reserved.
 */

#include "BasicOperator.h"
#include <iostream>
#include <sstream>
#include <iomanip>

using namespace std;

void CBasicOperator::FormatToken()
{
	CheckToken(); // Check if token is still present
	CK_RV rv;
	vector<CK_UTF8CHAR> pin;
	CK_UTF8CHAR label[MAX_LABEL_SIZE];
	// Get pin from user
	AskForPin(pin);
	// Get label from user
	AskForLabel(label, MAX_LABEL_SIZE);
	// Format token
	rv = mpFunctionList->C_InitToken(mCurrentSlot, &pin[0], pin.size(), label);
	if(CKR_OK != rv)
	{
		throw CPKCSException(rv, "CTokenOperator::FormatToken");
	}
}

void CBasicOperator::ChangePin()
{
	CheckToken(); // Check if token is still present
	// Get current PIN from user
	cout << "Obtaining existing PIN...\r\n";
	vector<CK_UTF8CHAR> pin;
	AskForPin(pin);

	while(!Authenticate(pin, CKS_RW_USER_FUNCTIONS))
	{
		cout << "Incorrect PIN. Please try again.\r\n";
	}
	CK_RV rv;

	// Get new PIN from user
	cout << "Obtaining new PIN...\r\n";
	vector<CK_UTF8CHAR> new_pin;
	AskForPin(new_pin);
	
	// Replace token's current PIN with new PIN
	rv = mpFunctionList->C_SetPIN(mCurrentSession, &pin[0], pin.size(),
		&new_pin[0], new_pin.size());
	if(CKR_OK != rv)
	{
		throw CPKCSException(rv, "CTokenOperator::ChangePin");
	}

	// Logout
	rv = mpFunctionList->C_Logout(mCurrentSession);
	if(CKR_OK != rv)
	{
		throw CPKCSException(rv, "CTokenOperator::ChangePin");
	}
}

void CBasicOperator::GetObjects()
{
	CheckToken(); // Check if token is still present
	Authenticate();
	// Get list of objects from token
	CK_RV rv;

	// Fill out object search template
	CK_OBJECT_HANDLE h_object;
	CK_ULONG ul_object_count;
	vector<CK_OBJECT_HANDLE> object_list;
	CK_ATTRIBUTE attr_template[] = {
		{CKA_CLASS, NULL_PTR, 0},
		{CKA_ID, NULL_PTR, 0},
		{CKA_LABEL, NULL_PTR, 0}
	};
	CK_ULONG attr_template_size = 3;

	// Initialize search
	rv = mpFunctionList->C_FindObjectsInit(mCurrentSession, NULL_PTR, 0);
	if(CKR_OK != rv)
	{
		throw CPKCSException(rv, "CTokenOperator::GetObjects");
	}
	// Get object handles
	rv = mpFunctionList->C_FindObjects(mCurrentSession, &h_object, 1,
		&ul_object_count);
	while(CKR_OK == rv && 0 != ul_object_count)
	{
		object_list.push_back(h_object);
		rv = mpFunctionList->C_FindObjects(mCurrentSession, &h_object, 1,
			&ul_object_count);
	} 
	if(CKR_OK != rv)
	{
		mpFunctionList->C_FindObjectsFinal(mCurrentSession);
		throw CPKCSException(rv, "CTokenOperator::GetObjects");
	}
	// Terminate search
	rv = mpFunctionList->C_FindObjectsFinal(mCurrentSession);

	// Get object labels
	vector<string> labels;
	for(ULONG i = 0; i < object_list.size(); i++)
	{
		CK_ULONG label_length = 0;

		// Set up attribute template for label
		CK_ATTRIBUTE tmplt = {CKA_LABEL, NULL_PTR, label_length};

		// Get label length
		rv = mpFunctionList->C_GetAttributeValue(mCurrentSession, object_list[i],
			&tmplt, 1);
		if(CKR_OK != rv)
		{
			throw CPKCSException(rv, "CTokenOperator::GetObjects");
		}

		// Get label
		string label;
		label.resize(tmplt.ulValueLen);
		tmplt.pValue = &label[0];
		rv = mpFunctionList->C_GetAttributeValue(mCurrentSession, object_list[i],
			&tmplt, 1);
		if(CKR_OK != rv)
		{
			throw CPKCSException(rv, "CTokenOperator::GetObjects");
		}

		// Store label in vector
		labels.push_back(label);
	}
	
	// Get object ids
	vector<string> ids;
	for(ULONG i = 0; i < object_list.size(); i++)
	{
		CK_ULONG id_length = 0;

		// Set up attribute template for object id
		CK_ATTRIBUTE tmplt = {CKA_ID, NULL_PTR, id_length};

		// Get object id length
		rv = mpFunctionList->C_GetAttributeValue(mCurrentSession, object_list[i],
			&tmplt, 1);
		if(CKR_OK != rv)
		{
			throw CPKCSException(rv, "CTokenOperator::GetObjects");
		}

		// Get object id
		vector<CK_BYTE> id(tmplt.ulValueLen);
		tmplt.pValue = &id[0];
		rv = mpFunctionList->C_GetAttributeValue(mCurrentSession, object_list[i],
			&tmplt, 1);
		if(CKR_OK != rv)
		{
			throw CPKCSException(rv, "CTokenOperator::GetObjects");
		}

		// Format object id as string
		stringstream id_str;
		for(size_t i = 0; i < id.size(); i++)
		{
			id_str << setw(2) << setfill('0') << hex
				   << static_cast<int>(id[i]);
		}

		// Store object id in vector
		ids.push_back(id_str.str());
	}

	// Get object types
	vector<string> types;
	for(ULONG i = 0; i < object_list.size(); i++)
	{
		CK_ULONG class_length = 0;

		// Set up attribute template for object type
		CK_ATTRIBUTE tmplt = {CKA_CLASS, NULL_PTR, class_length};

		// Get object type length
		rv = mpFunctionList->C_GetAttributeValue(mCurrentSession, object_list[i],
			&tmplt, 1);
		if(CKR_OK != rv)
		{
			throw CPKCSException(rv, "CTokenOperator::GetObjects");
		}

		// Get object type
		vector<CK_BYTE> type(tmplt.ulValueLen);
		tmplt.pValue = &type[0];
		rv = mpFunctionList->C_GetAttributeValue(mCurrentSession, object_list[i],
			&tmplt, 1);
		if(CKR_OK != rv)
		{
			throw CPKCSException(rv, "CTokenOperator::GetObjects");
		}
		
		// Format object type as string
		string type_str;
		CK_OBJECT_CLASS obj_class =
			*(reinterpret_cast<CK_OBJECT_CLASS_PTR>(&type[0]));
		switch(obj_class)
		{
			case CKO_DATA:
				type_str = "CKO_DATA";
				break;
			case CKO_CERTIFICATE:
				type_str = "CKO_CERTIFICATE";
				break;
			case CKO_PUBLIC_KEY:
				type_str = "CKO_PUBLIC_KEY";
				break;
			case CKO_PRIVATE_KEY:
				type_str = "CKO_PRIVATE_KEY";
				break;
			case CKO_SECRET_KEY:
				type_str = "CKO_SECRET_KEY";
				break;
			default:
				type_str = "other";
				break;
		}

		// Store object type in vector
		types.push_back(type_str);
	}

	// Print object attributes
	cout << "Token contains " << object_list.size() << " objects.\r\n";
	for(size_t i = 0; i < object_list.size(); i++)
	{
		cout << "Object " << i         << "\r\n" 
			 << "ID: "    << ids[i]    << "\r\n"
			 << "Label: " << labels[i] << "\r\n"
			 << "Type: "  << types[i]  << "\r\n\r\n";
	}

	// Logout from token
	rv = mpFunctionList->C_Logout(mCurrentSession);
	if(CKR_OK != rv)
	{
		throw CPKCSException(rv, "CTokenOperator::GetObjects");
	}
}

void CBasicOperator::AskForLabel(CK_UTF8CHAR_PTR pBuffer,
								 const CK_ULONG ulBufSize)
{
	// Fill buffer with blank characters (spaces)
	memset(pBuffer, 0x20, ulBufSize);
	
	string label;
	
	// Get label from user
	for(bool is_valid_label = false; !is_valid_label; )
	{
		cout << "Enter token label: ";
		cin >> label;
		if(label.length() <= 32 && !label.empty())
		{
			is_valid_label = true;
		}
	}
	// Copy label without null-terminator into buffer
	errno_t code = memcpy_s(pBuffer, ulBufSize, label.c_str(), label.length());
	if(0 != code)
	{
		const unsigned int buf_size = 256;
		char buffer[buf_size];
		_itoa_s(code,buffer,buf_size,10);
		string message("CTokenOperator::AskForLabel : "
					   "memcpy_s failed : code ");
		message += buffer;
		throw std::exception(message.c_str());
	}
}
