/**
 *  ESMART Token v3.1 - PKCS#11 Sample
 *
 *  Description: CTokenOperator realization
 *
 *  $Author: Alexander Revo
 *  $Date: 14.04.2014
 *  $Id: 3513-01-011
 *
 *  Copyright 2014 by ISBC Group
 *  All rights reserved.
 */

#include <Shlobj.h>
#include "TokenOperator.h"
#include <iostream>
#include <fstream>

using namespace std;

CTokenOperator::CTokenOperator()
{
	// Load %SYSTEMX86%/isbc_pkcs11_main.dll
	LoadPkcsDll();
	try
	{
		InitPkcsFunctionList(); // Get function pointers from DLL
		
		// Initialize PKCS library
		CK_RV rv = mpFunctionList->C_Initialize(NULL_PTR);
		if(CKR_OK != rv)
		{
			throw CPKCSException(rv, "CTokenOperator::CTokenOperator");
		}
	}
	catch(...)
	{
		// Unload DLL
		FreeLibrary(mPkcsDll);
		throw;
	}
}

CTokenOperator::~CTokenOperator()
{
	// Close all opened sessions, if any
	mpFunctionList->C_CloseAllSessions(mCurrentSlot);
	// Finish working with PKCS library
	mpFunctionList->C_Finalize(NULL_PTR);
	// Unload DLL
	FreeLibrary(mPkcsDll);
}

void CTokenOperator::Authenticate(CK_FLAGS flags)
{
	CheckToken(); // Check if token is still present
	std::vector<CK_UTF8CHAR> pin;
	// Get pin from user
	AskForPin(pin);
	// Keep asking
	while(!Authenticate(pin, flags))
	{
		cout << "Incorrect PIN. Please try again.\r\n";
		AskForPin(pin);
	}
}

bool CTokenOperator::Authenticate(const std::vector<CK_UTF8CHAR>& rPin,
								  CK_FLAGS flags)
{
	CheckToken(); // Check if token is still present
	bool result = false;

	CK_RV rv;
	// Open a new session
	rv = mpFunctionList->C_OpenSession(mCurrentSlot,CKF_SERIAL_SESSION | flags,
		NULL_PTR, NULL_PTR, &mCurrentSession);
	if(CKR_OK != NULL_PTR)
	{
		throw CPKCSException(rv, "CTokenOperator::Authenticate");
	}
	// Login into token as user
	rv = mpFunctionList->C_Login(mCurrentSession, CKU_USER,
		const_cast<CK_UTF8CHAR_PTR>(&rPin[0]), rPin.size());
	if(CKR_OK == rv)
	{
		result = true; // if C_Login is successful - return 'true'
	}
	else
	{
		// if incorrect PIN was entered - return 'false'
		if(CKR_PIN_INCORRECT != rv) 
		{
			// if any other error - throw exception
			throw CPKCSException(rv, "CTokenOperator::Authenticate"); 
		}
	}
	return result;
}

std::wstring CTokenOperator::GetLibraryPath()
{
	wchar_t* buffer;
	wstring result;

	// Get path to %WinDir%\System32 on x86 system or %WinDir%\SysWOW64 on x64
	SHGetKnownFolderPath(FOLDERID_SystemX86, 0, 0, &buffer);
	try
	{
		result.assign(buffer);
	}
	catch(...)
	{
		// Free memory allocated by SHGetKnownFolderPath()
		CoTaskMemFree(buffer);
		throw;
	}
	CoTaskMemFree(buffer); // Free memory allocated by SHGetKnownFolderPath()

	result += L"\\isbc_pkcs11_main.dll";

	return result;
}

void CTokenOperator::LoadPkcsDll()
{
	// Get DLL path and try to load it
	mPkcsDll = LoadLibraryW(GetLibraryPath().c_str());
	if(NULL == mPkcsDll)
	{
		throw std::exception("could not load PKCS library");
	}
}

void CTokenOperator::InitPkcsFunctionList()
{
	// Get function list from DLL
	CK_RV (*func)(CK_FUNCTION_LIST_PTR_PTR);
	func = (CK_RV (__cdecl *)(CK_FUNCTION_LIST_PTR_PTR))GetProcAddress(
		mPkcsDll, "C_GetFunctionList");

	CK_RV rv = (*func)(&mpFunctionList);
	if(CKR_OK != rv)
	{
		throw CPKCSException(rv, "CTokenOperator::InitPkcsFunctionList");
	}
}

bool CTokenOperator::SelectSlot()
{
	// Get list of slots with token inserted
	vector<CK_SLOT_ID> slots_with_tokens = GetSlotsWithToken();
	if(slots_with_tokens.empty())
	{
		return false;
	}

	// If such slots exist, ask user to select one
	if(slots_with_tokens.size()>1)
	{
		mCurrentSlot = AskForSlot(slots_with_tokens);
	}
	else
	{
		mCurrentSlot = slots_with_tokens[0];
	}
	cout << "Selected slot: " << GetSlotName(mCurrentSlot) << "\r\n";

	return true;
}

std::vector<CK_SLOT_ID> CTokenOperator::GetSlotsWithToken()
{
	CK_RV rv;
	
	// Get number of slots with token inserted
	CK_ULONG ul_slot_with_token_count;
	rv = mpFunctionList->C_GetSlotList(CK_TRUE, NULL_PTR,
		&ul_slot_with_token_count);
	if(CKR_OK != rv)
	{
		throw CPKCSException(rv, "CTokenOperator::GetSlotsWithToken");		
	}

	std::vector<CK_SLOT_ID> result;
	// If there are slots with token,
	if(0 != ul_slot_with_token_count)
	{
		// get IDs of slots with token inserted
		result.resize(ul_slot_with_token_count);
		rv = mpFunctionList->C_GetSlotList(CK_TRUE, &result[0],
			&ul_slot_with_token_count);
		if(CKR_OK != rv)
		{
			throw CPKCSException(rv, "CTokenOperator::GetSlotsWithToken");		
		}
	}
	return result;
}

CK_SLOT_ID CTokenOperator::AskForSlot(const std::vector<CK_SLOT_ID>& rSlots)
{
	// If no slots with tokens are available - throw exception
	if(0 == rSlots.size())
	{
		throw std::exception("CTokenOperator::AskForSlot :"
							 " no tokens inserted (slots.size() == 0)");
	}
	CK_SLOT_ID result = 0;
	bool is_valid_choice = false;
	
	// Print list of slots with tokens
	cout << "Slots with tokens:\r\n";
	for(UINT i = 0; i < rSlots.size(); i++)
	{
		cout << "[" << i << "] : " << GetSlotName(rSlots[i]) << endl;
	}
	// Ask user to enter slot number
	while(!is_valid_choice)
	{
		cout << "\r\nPlease enter slot number: ";

		UINT choice = 0;
		cin >> choice;

		if(choice < rSlots.size())
		{
			is_valid_choice = true;
			result = rSlots[choice];
		}
	}

	return result;
}

std::string CTokenOperator::GetSlotName(const CK_SLOT_ID slotId)
{
	string result;

	CK_RV rv;
	CK_SLOT_INFO slot_info;

	// Get slot info
	rv = mpFunctionList->C_GetSlotInfo(slotId, &slot_info);
	if (CKR_OK != rv)
	{
		throw CPKCSException(rv, "CTokenOperator::GetSlotName");		
	}
	result = reinterpret_cast<char*>(slot_info.slotDescription);

	// Remove extra characters at the end
	result.resize(sizeof(slot_info.slotDescription));
	int i = result.find_last_of(0x20);
	while(string::npos != i && 0x20 == result[i])
	{
		i--;
	}
	result.resize(++i);

	return result;
}

void CTokenOperator::AskForPin(std::vector<CK_UTF8CHAR>& rBuffer)
{
	// Get PIN from user
	string pin;
	bool is_valid_pin = false;
	do
	{
		cout << "Enter PIN: ";
		cin >> pin;
		if(!pin.empty())
		{
			is_valid_pin = true;
		}
	} while(!is_valid_pin);

	// Copy string contents without null-terminator into buffer
	rBuffer.assign(pin.begin(), pin.end());
}

void CTokenOperator::CheckToken()
{
	// Attempt to get token info from current slot
	CK_RV rv;
	CK_TOKEN_INFO tmp;
	rv = mpFunctionList->C_GetTokenInfo(mCurrentSlot, &tmp);

	// If failed - throw exception
	if(CKR_OK != rv)
	{
		throw CPKCSException(rv, "CTokenOperator::CheckToken");		
	}
}
